#include "ehs/io/hid/HID.h"

#include "ehs/Log.h"
#include "ehs/Math.h"
#include "ehs/io/hid/Keyboard.h"

namespace ehs
{
	HID::HID()
		: type(EHS_HID_UNKNOWN), hashName(0), id(0), heldTime(0.0), activateTime(0.0f), active(false)
	{
	}

	HID::HID(const UInt_8 type, Str_8 name, const UInt_64 id)
		: type(type), hashName(name.Hash_64()), name((Str_8 &&)name), id(id), heldTime(0.0f), activateTime(0.0f),
		active(false)
	{
	}

	HID::HID(HID&& hid) noexcept
		: type(hid.type), hashName(hid.hashName), name((Str_8&&)hid.name), id(hid.id),
		states((Array<ButtonState> &&)hid.states), lastState((Button &&)hid.lastState), heldTime(hid.heldTime),
		activateTime(hid.activateTime), active(hid.active)
	{
		hid.type = EHS_HID_UNKNOWN;
		hid.hashName = 0;
		hid.id = 0;
		hid.lastState = {};
		hid.heldTime = 0.0f;
		hid.activateTime = 0.0f;
		hid.active = false;
	}

	HID::HID(const HID& hid)
		: type(hid.type), hashName(hid.hashName), name(hid.name), id(hid.id), states(hid.states), heldTime(0.0f),
		activateTime(0.0f), active(false)
	{
	}

	HID& HID::operator=(HID&& hid) noexcept
	{
		if (this == &hid)
			return *this;

		type = hid.type;
		hashName = hid.hashName;
		name = (Str_8&&)hid.name;
		id = hid.id;
		states = (Array<ButtonState> &&)hid.states;
		lastState = (Button &&)hid.lastState;
		heldTime = hid.heldTime;
		activateTime = hid.activateTime;
		active = hid.active;

		hid.type = EHS_HID_UNKNOWN;
		hid.hashName = 0;
		hid.id = 0;
		hid.lastState = {};
		hid.heldTime = 0.0f;
		hid.activateTime = 0.0f;
		hid.active = false;

		return *this;
	}

	HID& HID::operator=(const HID& hid)
	{
		if (this == &hid)
			return *this;

		type = hid.type;
		hashName = hid.hashName;
		name = hid.name;
		id = hid.id;
		states = hid.states;
		lastState = {};
		heldTime = 0.0f;
		activateTime = 0.0f;
		active = false;

		return *this;
	}

	bool HID::operator==(const HID& other) const
	{
		return type == other.type && hashName == other.hashName;
	}

	bool HID::operator!=(const HID& other) const
	{
		return type != other.type || hashName != other.hashName;
	}

	bool HID::operator==(const UInt_64 otherId) const
	{
		return id == otherId;
	}

	bool HID::operator!=(const UInt_64 otherId) const
	{
		return id != otherId;
	}

	void HID::Poll(const float delta)
	{
		for (UInt_64 i = 0; i < states.Size(); ++i)
		{
			if (states[i].IsPressed())
			{
				if (states[i].GetState() == State::RELEASED)
				{
					states[i].SetState(State::TOUCHED);

					lastState = states[i].GetButton();
					ResetTime();
					active = true;
				}
				else
				{
					states[i].SetState(State::PRESSED);

					if (lastState == states[i].GetButton())
					{
						if (TickHoldTime(delta))
							TickActivateTime(delta);
					}
				}
			}
			else
			{
				if (states[i].GetState() == State::PRESSED || states[i].GetState() == State::TOUCHED)
					states[i].SetState(State::JUST_RELEASED);
				else
					states[i].SetState(State::RELEASED);

				if (lastState == states[i].GetButton())
				{
					lastState = {};
					ResetTime();
					active = false;
				}
			}
		}
	}

	UInt_8 HID::GetType() const
	{
		return type;
	}

	Str_8 HID::GetName() const
	{
		return name;
	}

	UInt_64 HID::GetId() const
	{
		return id;
	}

	void HID::ReleaseAll()
	{
		for (UInt_64 i = 0; i < states.Size(); ++i)
			states[i].SetPressed(false);
	}

	Vector<const ButtonState*> HID::GetAllTouched() const
	{
		Vector<const ButtonState*> result(0, states.Size() + 1);

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::TOUCHED)
				result.Push(&states[i]);

		return result;
	}

	const ButtonState* HID::IsTouched(const Button& button) const
	{
		const ButtonState* state = GetState(button);
		if (!state || *state != State::TOUCHED)
			return nullptr;

		return state;
	}

	const ButtonState* HID::IsTouched() const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::TOUCHED)
				return &states[i];

		return nullptr;
	}

	Vector<const ButtonState*> HID::GetAllDown() const
	{
		Vector<const ButtonState*> result(0, states.Size() + 1);

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::PRESSED)
				result.Push(&states[i]);

		return result;
	}

	const ButtonState* HID::IsDown(const Button& button) const
	{
		const ButtonState* state = GetState(button);
		if (!state || (*state != State::PRESSED))
			return nullptr;

		return state;
	}

	const ButtonState* HID::IsDown() const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::PRESSED)
				return &states[i];

		return nullptr;
	}

	Vector<const ButtonState*> HID::GetAllJustReleased() const
	{
		Vector<const ButtonState*> result(0, states.Size() + 1);

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::JUST_RELEASED)
				result.Push(&states[i]);

		return result;
	}

	const ButtonState* HID::IsJustReleased(const Button& button) const
	{
		const ButtonState* state = GetState(button);
		if (!state || *state != State::JUST_RELEASED)
			return nullptr;

		return state;
	}

	const ButtonState* HID::IsJustReleased() const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::JUST_RELEASED)
				return &states[i];

		return nullptr;
	}

	Vector<const ButtonState*> HID::GetAllUp() const
	{
		Vector<const ButtonState*> result(0, states.Size() + 1);

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::RELEASED || states[i] == State::JUST_RELEASED)
				result.Push(&states[i]);

		return result;
	}

	const ButtonState* HID::IsUp(const Button& button) const
	{
		const ButtonState* state = GetState(button);
		if (!state || (*state != State::RELEASED && *state != State::JUST_RELEASED))
			return nullptr;

		return state;
	}

	const ButtonState* HID::IsUp() const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::RELEASED || states[i] == State::JUST_RELEASED)
				return &states[i];

		return nullptr;
	}

	const ButtonState *HID::IsPressed(const Button &button)
	{
		if (active && lastState == button)
			return GetState(lastState);

		return nullptr;
	}

	const ButtonState *HID::GetPressed()
	{
		if (active)
			return GetState(lastState);

		return nullptr;
	}

	void HID::ButtonDown(const Button& button)
	{
		if (ButtonState* state = GetState(button); state)
			state->SetPressed(true);
		else
			states.Push(ButtonState(button, State::RELEASED));
	}

	void HID::ButtonUp(const Button& button)
	{
		if (ButtonState* state = GetState(button); state)
			state->SetPressed(false);
		else
			states.Push(ButtonState(button, State::JUST_RELEASED));
	}

	const ButtonState* HID::GetState(const Button& button) const
	{
		if (!button.IsValid())
			return nullptr;

		for (UInt_64 i = 0; i < states.Size(); ++i)
			if (states[i] == button)
				return &states[i];

		return nullptr;
	}

	bool HID::IsValid() const
	{
		return id;
	}

	HID* HID::Clone() const
	{
		return new HID(*this);
	}

	bool HID::HasState(const Button& button) const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i].GetButton() == button)
				return true;

		return false;
	}

	bool HID::AddState(const ButtonState& state)
	{
		if (HasState(state.GetButton()))
			return false;

		states.Push(state);

		return true;
	}

	ButtonState* HID::GetState(const Button& button)
	{
		if (!button.IsValid())
			return nullptr;

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i].GetButton() == button)
				return &states[i];

		return nullptr;
	}

	bool HID::TickHoldTime(const float delta)
	{
		if ((heldTime += delta) > 0.5f)
			return true;

		active = false;

		return false;
	}

	void HID::ResetTime()
	{
		heldTime = 0.0f;
		activateTime = 0.0f;
	}

	void HID::TickActivateTime(const float delta)
	{
		if ((activateTime += delta) > 0.075f)
		{
			activateTime = ehs::Math::Mod(activateTime, 0.075f);
			active = true;
		}
		else
		{
			active = false;
		}
	}
}
